# taken from https://github.com/lopeLH/Multilayer-Perceptron
# with slight modifications
__author__ = 'lope'

import numpy as np
import math


def softmax(x):
    if x.ndim == 1:
        e = np.exp(x - np.max(x))  # prevent overflow
        return e / (np.sum(e, axis=0))
    else:
        e = np.exp(x - np.max(x,axis=1,keepdims=True))
        return e / (np.array([np.sum(e, axis=1)]).T)  # ndim = 2


class MLP:

#   def sigmoid(self, x):
#       return float(1.0 / (1.0 + math.exp(-x)))

    def __init__(self,nInput,nHidden,nOutput):
        self.WList = [np.random.rand(nHidden, nInput+1), np.random.rand(nOutput, nHidden+1)]
        self.sizeList = [ np.prod(w.shape) for w in self.WList]
        self.nInput = nInput;
        self.nHidden = nHidden;
        self.nOutput = nOutput;

        # self.sigmoid = np.vectorize(self.sigmoid)

    def getParams(self):
        return np.hstack((self.WList[0].reshape(-1),self.WList[1].reshape(-1)))

    def setParams(self, params):
        assert(len(params) == sum(self.sizeList))
        idx = 0
        for i,s in enumerate(self.sizeList):
            self.WList[i] = params[idx:idx+s].reshape(self.WList[i].shape)
            idx+=s

    def compute(self, inputValues):
        inputValues = np.append([1], inputValues)
        self.lastInput = inputValues;
        self.lastHiddenOutput = np.dot(self.WList[0], np.transpose(inputValues))
        self.lastHiddenOutput = np.tanh(self.lastHiddenOutput)
        self.lastHiddenOutput  = np.append([1], self.lastHiddenOutput)
        self.lastOutput = np.dot(self.WList[1], self.lastHiddenOutput)
        # self.lastOutput = self.sigmoid(self.lastOutput)
        return self.lastOutput

class Backpropagation:

    def __init__(self, mlp, alpha, regularization):
        self.DeltaList = [np.zeros((mlp.nHidden, mlp.nInput+1)), np.zeros((mlp.nOutput, mlp.nHidden+1))]

        self.alpha = alpha;
        self.regularization = regularization;
        self.mlp = mlp


    def iterate(self, inputValues, outputValues):
        self.DeltaList = [np.zeros((mlp.nHidden, mlp.nInput+1)), np.zeros((mlp.nOutput, mlp.nHidden+1))]

        for i in range(len(inputValues)):
            self.mlp.compute(inputValues[i])
            delta3 = np.subtract( self.mlp.lastOutput, outputValues[i]);

            Od3 = np.dot((self.mlp.WList[1].transpose()),delta3)
            gz2 = (1.0-np.square(self.mlp.lastHiddenOutput))
            delta2 = np.multiply(Od3, gz2)

            self.DeltaList[1] = np.add(self.DeltaList[1], np.outer(delta3, self.mlp.lastHiddenOutput))
            self.DeltaList[0] = np.add(self.DeltaList[0], np.delete(np.outer(delta2, self.mlp.lastInput),0,0))

        self.DeltaList[0]   = self.DeltaList[0]/len(inputValues) + self.regularization*self.mlp.WList[0]
        self.DeltaList[1]   = self.DeltaList[1]/len(inputValues) + self.regularization*self.mlp.WList[1]
        self.mlp.WList[0] = self.mlp.WList[0]-self.alpha*self.DeltaList[0]
        self.mlp.WList[1] = self.mlp.WList[1]-self.alpha*self.DeltaList[1]
